{----------------------------------------------------------------------------}
{ INVASORES                                                                  }
{                                                                            }
{ Escrito por Marco Antonio Alvarado                                         }
{ Costa Rica, 1996.                                                          }

{$R-}   {Sin revisin de rango.}
{$Q-}

PROGRAM Invasores;

  USES
    CRT;

  TYPE
    TColeccion = ARRAY [0..0] OF Byte;
    PColeccion = ^TColeccion;

  VAR
    ViejaInterrupcionTeclado : Pointer;
    ViejaInterrupcionReloj : Pointer;

  PROCEDURE PoneInterrupcion (IndiceInterrupcion : Byte; Direccion : Pointer);
  BEGIN

    ASM
{Pascal obliga a salvar los registros BP, SS, SP y DS.}
      PUSH DS

{Pone la interrupcin.}
      LDS DX, Direccion
      MOV AL, IndiceInterrupcion
      MOV AH, 37
      INT 33

{Pascal obliga a reponer los registros BP, SS, SP y DS.}
      POP DS
    END;

  END;

  PROCEDURE TomaInterrupcion (IndiceInterrupcion : Byte; VAR Direccion :
    Pointer);

    VAR
      Segmento : Word;
      Desplazamiento : Word;

  BEGIN

    ASM
{Toma la interrupcin.}
      MOV AL, IndiceInterrupcion
      MOV AH, 53
      INT 33
      MOV Segmento, ES
      MOV Desplazamiento, BX
    END;

    Direccion := Ptr (Segmento, Desplazamiento);
  END;

{----------------------------------------------------------------------------}
{ TECLADO                                                                    }

  VAR
    Tecla : Byte;
    TeclaExtendida : Boolean;
    TeclaEsc : Boolean;
    TeclaBarra : Boolean;
    TeclaArriba : Boolean;
    TeclaIzquierda : Boolean;
    TeclaDerecha : Boolean;
    TeclaAbajo : Boolean;

{$F+}
  PROCEDURE LeeTecla; INTERRUPT;
  BEGIN
{Obtiene el cdigo de la tecla.}
    Tecla := Port [96];

    IF TeclaExtendida THEN
    BEGIN
      TeclaExtendida := False;

      CASE Tecla OF
        72 : TeclaArriba := True;
        75 : TeclaIzquierda := True;
        77 : TeclaDerecha := True;
        80 : TeclaAbajo := True;
        200 : TeclaArriba := False;
        203 : TeclaIzquierda := False;
        205 : TeclaDerecha := False;
        208 : TeclaAbajo := False;
      END;

    END ELSE
    BEGIN

      CASE Tecla OF
        1 : TeclaEsc := True;
        57 : TeclaBarra := True;
        129 : TeclaEsc := False;
        185 : TeclaBarra := False;
        224 : TeclaExtendida := True;
      END;

    END;

{Reconoce la interrupcin.}
    Port [32] := 32;
  END;
{$F-}

  PROCEDURE PreparaTeclado;
  BEGIN
    TeclaEsc := False;
    TeclaBarra := False;
    TeclaArriba := False;
    TeclaIzquierda := False;
    TeclaDerecha := False;
    TeclaAbajo := False;
    TeclaExtendida := False;
    TomaInterrupcion (9, ViejaInterrupcionTeclado);
    PoneInterrupcion (9, @LeeTecla);
  END;

  PROCEDURE TerminaTeclado;
  BEGIN
    PoneInterrupcion (9, ViejaInterrupcionTeclado);
  END;

{ TECLADO                                                                    }
{----------------------------------------------------------------------------}

{----------------------------------------------------------------------------}
{ VIDEO                                                                      }

  TYPE

    TCuadro = RECORD
                Anchura : Word;
                Altura : Word;
                Coleccion : PColeccion;
              END;

    TPaleta = RECORD
                Coleccion : PColeccion;
              END;


  VAR
    ModoVideo : Byte ABSOLUTE $40:$49;
    ViejoModoVideo : Byte;
    Pantalla : PColeccion;
    Paleta : TPaleta;
    MosaicoNaves : TCuadro;

  PROCEDURE PreparaCuadro (VAR Cuadro : TCuadro; NuevaAnchura, NuevaAltura :
    Word);
  BEGIN

    WITH Cuadro DO
    BEGIN
      Anchura := NuevaAnchura;
      Altura := NuevaAltura;
      GetMem (Coleccion, Anchura * Altura);
    END;

  END;

  PROCEDURE TerminaCuadro (VAR Cuadro : TCuadro);
  BEGIN

    WITH Cuadro DO
    BEGIN
      FreeMem (Coleccion, Anchura * Altura);
    END;

  END;

  PROCEDURE PreparaPaleta (VAR Paleta : TPaleta);
  BEGIN

    WITH Paleta DO
    BEGIN
      GetMem (Coleccion, 768);
    END;

  END;

  PROCEDURE TerminaPaleta (VAR Paleta : TPaleta);
  BEGIN

    WITH Paleta DO
    BEGIN
      FreeMem (Coleccion, 768);
    END;

  END;

  PROCEDURE CargaPaletaPCX (NombrePCX : STRING; Destino : TPaleta);

    VAR
      PCX : FILE;
      IndiceIntensidad : Word;

  BEGIN

    WITH Destino DO
    BEGIN
{Carga toda la paleta en el final del archivo.}
      Assign (PCX, NombrePCX);
      Reset (PCX, 1);
      Seek (PCX, FileSize (PCX) - 768);
      BlockRead (PCX, Coleccion^, 768);
      Close (PCX);

{Ajusta las intensidades RVA.}
      FOR IndiceIntensidad := 0 TO 767 DO
        Coleccion^ [IndiceIntensidad] := Coleccion^ [IndiceIntensidad] SHR 2;
    END;

  END;

  PROCEDURE CargaCuadroPCX (NombrePCX : STRING; Destino : TCuadro);

    VAR
      PCX : FILE;
      TamanoComprimido : Word;
      Comprimido : PColeccion;
      Tamano : Word;
      Cuenta : Word;
      IndiceComprimido : Word;
      Corrida : Word;
      IndiceColeccion : Word;

  BEGIN
{Carga todo el cuadro comprimido a la memoria.}
    Assign (PCX, NombrePCX);
    Reset (PCX, 1);
    Seek (PCX, 128);
    TamanoComprimido := FileSize (PCX) - 897;
    GetMem (Comprimido, TamanoComprimido);
    BlockRead (PCX, Comprimido^, TamanoComprimido);
    Close (PCX);

{Descomprime.}
    Cuenta := 0;
    IndiceComprimido := 0;
    IndiceColeccion := 0;

    WITH Destino DO
    BEGIN
      Tamano := Anchura * Altura;

      WHILE Cuenta < Tamano DO
      BEGIN

        IF (Comprimido^ [IndiceComprimido] AND 192) = 192 THEN
        BEGIN
          Corrida := Comprimido^ [IndiceComprimido] AND 63;
          Inc (IndiceComprimido);
          FillChar (Coleccion^ [IndiceColeccion], Corrida, Comprimido^
            [IndiceComprimido]);
          Inc (IndiceComprimido);
          Inc (IndiceColeccion, Corrida);
          Inc (Cuenta, Corrida);
        END ELSE
        BEGIN
          Coleccion^ [IndiceColeccion] := Comprimido^ [IndiceComprimido];
          Inc (IndiceComprimido);
          Inc (IndiceColeccion);
          Inc (Cuenta);
        END;

      END;

    END;

{Libera la memoria del cuadro comprimido.}
    FreeMem (Comprimido, TamanoComprimido);
  END;

  PROCEDURE EsperaRefrescamiento;

    LABEL
      Espera1, Espera2;

  BEGIN

    ASM
      MOV DX, 986

{Espera que se refresque la pantalla.}
      Espera1:

      IN AL, DX
      AND AL, 8
      JNZ Espera1

{Espera mientras se refresca la pantalla.}
      Espera2:

      IN AL, DX
      AND AL, 8
      JZ Espera1
    END;

  END;

  PROCEDURE PoneColor (IndiceColor, R, V, A : Byte);
  BEGIN

    ASM
{Pone las intensidades RVA del color IndiceColor.}
      MOV DX, 968
      MOV AL, IndiceColor
      OUT DX, AL
      INC DX
      MOV AL, R
      OUT DX, AL
      MOV AL, V
      OUT DX, AL
      MOV AL, A
      OUT DX, AL
    END;

  END;

  PROCEDURE TomaColor (IndiceColor : Byte; VAR R, V, A : Byte);

    VAR
      R1 : Byte;
      V1 : Byte;
      A1 : Byte;

  BEGIN

    ASM
{Toma las intensidades RVA del color IndiceColor.}
      MOV DX, 967
      MOV AL, IndiceColor
      OUT DX, AL
      INC DX
      INC DX
      IN AL, DX
      MOV R1, AL
      IN AL, DX
      MOV V1, AL
      IN AL, DX
      MOV A1, AL
    END;

    R := R1;
    V := V1;
    A := A1;
  END;

  PROCEDURE PonePaleta (Fuente : TPaleta);

    LABEL
      Ciclo;

  BEGIN
    EsperaRefrescamiento;

    ASM
{Pascal obliga a salvar los registros BP, SS, SP y DS.}
      PUSH DS

{Pone la paleta.}
      LDS SI, Fuente. Coleccion
      MOV CX, 768
      MOV DX, 968
      XOR AL, AL
      OUT DX, AL
      MOV DX, 969

      Ciclo:

      LODSB
      OUT DX, AL

      LOOP Ciclo

{Pascal obliga a reponer los registros BP, SS, SP y DS.}
      POP DS
    END;

  END;

  PROCEDURE TomaPaleta (Fuente : TPaleta);

    LABEL
      Ciclo;

  BEGIN

    ASM
{Toma la paleta.}
      LES DI, Fuente. Coleccion
      MOV CX, 768
      MOV DX, 967
      XOR AL, AL
      OUT DX, AL
      MOV DX, 969

      Ciclo:

      IN AL, DX
      STOSB

      LOOP Ciclo
    END;

  END;

  PROCEDURE PoneModoVideo (IndiceModo : Byte);
  BEGIN

    ASM
{Pone el modo de video.}
      XOR AH, AH
      MOV AL, IndiceModo
      INT 16
    END;

  END;

  PROCEDURE MuestraMosaico (Mosaico : TCuadro; Columna, Fila : Byte; X, Y :
    Word);

    VAR
      IndiceMosaico : Word;
      IndicePantalla : Word;
      IndiceLinea : Byte;

  BEGIN
    IndiceMosaico := Fila * 6400 + Columna * 20;
    IndicePantalla := Y * 320 + X;

    FOR IndiceLinea := 1 TO 19 DO
    BEGIN
      Move (Mosaico. Coleccion^ [IndiceMosaico], Pantalla^ [IndicePantalla],
        19);
      Inc (IndiceMosaico, 320);
      Inc (IndicePantalla, 320);
    END;

  END;

  PROCEDURE MuestraPunto (X, Y : Word; Color : Byte);
  BEGIN
    Pantalla^ [Y SHL 8 + Y SHL 6 + X] := Color;
  END;

  PROCEDURE PreparaVideo;
  BEGIN
    Pantalla := Ptr ($a000, 0);
    PreparaPaleta (Paleta);
    PreparaCuadro (MosaicoNaves, 320, 200);
    CargaPaletaPCX ('INVASOR.PCX', Paleta);
    CargaCuadroPCX ('INVASOR.PCX', MosaicoNaves);
    ViejoModoVideo := ModoVideo;
    PoneModoVideo (19);
    PonePaleta (Paleta);
  END;

  PROCEDURE TerminaVideo;
  BEGIN
    PoneModoVideo (ViejoModoVideo);
    TerminaCuadro (MosaicoNaves);
    TerminaPaleta (Paleta);
  END;

{ VIDEO                                                                      }
{----------------------------------------------------------------------------}

{----------------------------------------------------------------------------}
{ OBJETOS                                                                    }

  TYPE

    OBarrera = OBJECT
                 PROCEDURE Muestra;
               END;

    OBala = OBJECT
              X : Word;
              Y : Word;
              Volando : Boolean;
              Chocando : Boolean;
              Color : Byte;
              CambioY : ShortInt;
              CONSTRUCTOR Prepara (NuevoCambioY : ShortInt);
              PROCEDURE Transforma;
              PROCEDURE Muestra;
              DESTRUCTOR Termina;
            END;

    OArtilleria = OBJECT
                    Bala : OBala;
                    X : Word;
                    Y : Word;
                    Columna : Byte;
                    Disparando : Boolean;
                    Muriendo : Boolean;
                    Martillando : Boolean;
                    Muerta : Boolean;
                    CONSTRUCTOR Prepara;
                    PROCEDURE Transforma;
                    PROCEDURE DetectaChoque;
                    PROCEDURE Muestra;
                    DESTRUCTOR Termina;
                  END;

    OInvasor = OBJECT
                 Bala : OBala;
                 X : Word;
                 Y : Word;
                 Columna : Byte;
                 Fila : Byte;
                 Muriendo : Boolean;
                 Muerto : Boolean;
                 CONSTRUCTOR Prepara (NuevaX, NuevaY, NuevaFila : Byte);
                 PROCEDURE Transforma;
                 PROCEDURE DetectaChoque;
                 PROCEDURE Muestra;
                 DESTRUCTOR Termina;
               END;

    OGrupoInvasor = OBJECT
                      Invasores : ARRAY [1..10] OF OInvasor;
                      Muertos : Boolean;
                      CambioX : ShortInt;
                      CONSTRUCTOR Prepara;
                      PROCEDURE Transforma;
                      PROCEDURE DetectaChoque;
                      PROCEDURE Muestra;
                      DESTRUCTOR Termina;
                    END;

  VAR
    Barrera : OBarrera;
    Artilleria : OArtilleria;
    GrupoInvasor : OGrupoInvasor;

  PROCEDURE OBarrera. Muestra;
  BEGIN
    MuestraMosaico (MosaicoNaves, 0, 0, 150, 160);
    MuestraMosaico (MosaicoNaves, 0, 0, 90, 160);
    MuestraMosaico (MosaicoNaves, 0, 0, 210, 160);
    MuestraMosaico (MosaicoNaves, 0, 0, 30, 160);
    MuestraMosaico (MosaicoNaves, 0, 0, 270, 160);
  END;

  CONSTRUCTOR OBala. Prepara;
  BEGIN
    Volando := False;
    Color := 255;
    CambioY := NuevoCambioY;
  END;

  PROCEDURE OBala. Transforma;
  BEGIN

    IF Volando THEN
    BEGIN
      MuestraPunto (X, Y, 0);

      IF CambioY < 0 THEN
      BEGIN

        IF Y > 0 THEN
          Dec (Y)
        ELSE
          Volando := False;

      END ELSE
        IF CambioY > 0 THEN
        BEGIN

          IF Y < 200 THEN
            Inc (Y)
          ELSE
            Volando := False;

        END;

    END;

    IF (Y > 150) AND (Y < 180) THEN
      IF Pantalla^ [320 * Y + X] <> 0 THEN
      BEGIN
        MuestraPunto (X, Y, 0);
        Chocando := True;
        Volando := False;
      END;

  END;

  PROCEDURE OBala. Muestra;
  BEGIN
    IF Volando THEN
      MuestraPunto (X, Y, Color);
  END;

  DESTRUCTOR OBala. Termina;
  BEGIN
  END;

  CONSTRUCTOR OArtilleria. Prepara;
  BEGIN
    Bala. Prepara (-1);
    X := 150;
    Y := 180;
    Columna := 0;
    Disparando := False;
    Muriendo := False;
    Muerta := False;
  END;

  PROCEDURE OArtilleria. Transforma;
  BEGIN

    IF NOT Muriendo THEN
    BEGIN

      IF TeclaIzquierda XOR TeclaDerecha THEN
        IF TeclaIzquierda THEN
        BEGIN
          IF X > 0 THEN Dec (X);
        END ELSE
        BEGIN
          IF X < 301 THEN Inc (X);
        END;

      IF TeclaBarra AND (NOT Bala. Volando) THEN
      BEGIN
        Disparando := True;
        Martillando := True;
        Bala. X := X + 9;
        Bala. Y := Y;
        Bala. Volando := True;
      END;

      IF Disparando THEN
        IF Martillando THEN
        BEGIN

          CASE Columna OF
            0 : Columna := 1;
            1 : Columna := 2;
            2 : Martillando := False;
          END;

        END ELSE
        BEGIN

          CASE Columna OF
            2 : Columna := 1;
            1 : Columna := 0;
            0 : Disparando := False;
          END;

        END;

    END ELSE
    BEGIN

      CASE Columna OF
        0..2 : IF Random (10) = 9 THEN Columna := 15;
        15 : IF Random (10) = 9 THEN Columna := 14;
        14 : IF Random (10) = 9 THEN Columna := 13;
        13 : IF Random (10) = 9 THEN Columna := 12;
        12 : IF Random (10) = 9 THEN Columna := 11;
        11 : Muerta := True;
      END;

    END;

    Bala. Transforma;
  END;

  PROCEDURE OArtilleria. DetectaChoque;

    VAR
      IndiceX : Byte;

  BEGIN
    FOR IndiceX := 0 TO 18 DO
      IF Pantalla^ [320 * Y + X + IndiceX] <> 0 THEN Muriendo := True;
  END;

  PROCEDURE OArtilleria. Muestra;
  BEGIN
    MuestraMosaico (MosaicoNaves, Columna, 1, X, Y);
    Bala. Muestra;
  END;

  DESTRUCTOR OArtilleria. Termina;
  BEGIN
    Bala. Termina;
  END;

  CONSTRUCTOR OInvasor. Prepara;
  BEGIN
    X := NuevaX;
    Y := NuevaY;
    Fila := NuevaFila;
    Bala. Prepara (1);
    Muriendo := False;
    Muerto := False;
  END;

  PROCEDURE OInvasor. Transforma;

    VAR
      Posibilidad : Word;

  BEGIN

    IF NOT Muriendo THEN
    BEGIN
      Posibilidad := 1;
      IF X < Artilleria. X THEN
        Posibilidad := (Artilleria. X - X) * 5
      ELSE
        IF X > Artilleria. X THEN
          Posibilidad := (X - Artilleria. X) * 5;

      IF (NOT Bala. Volando) THEN
        IF Random (Posibilidad) = Posibilidad - 1 THEN
        BEGIN
          Bala. X := X + 9;
          Bala. Y := Y + 20;
          Bala. Volando := True;
        END;

      CASE Columna OF
        0 : IF Random (10) = 9 THEN Columna := 1;
        1 : IF Random (10) = 9 THEN Columna := 2;
        2 : IF Random (10) = 9 THEN Columna := 0;
      END;

    END ELSE
    BEGIN

      CASE Columna OF
        0..2 : IF Random (10) = 9 THEN Columna := 15;
        15 : IF Random (100) = 99 THEN Columna := 14;
        14 : IF Random (10) = 9 THEN Columna := 13;
        13 : IF Random (10) = 9 THEN Columna := 12;
        12 : IF Random (10) = 9 THEN Columna := 11;
        11 : IF Random (10) = 9 THEN Columna := 10;
        10 : Muerto := True;
      END;

    END;

    Bala. Transforma;
  END;

  PROCEDURE OInvasor. DetectaChoque;

    VAR
      IndiceX : Byte;

  BEGIN
    FOR IndiceX := 0 TO 18 DO
      IF Pantalla^ [320 * (Y + 19) + X + IndiceX] <> 0 THEN Muriendo := True;
  END;

  PROCEDURE OInvasor. Muestra;
  BEGIN
    MuestraMosaico (MosaicoNaves, Columna, Fila, X, Y);
    Bala. Muestra;
  END;

  DESTRUCTOR OInvasor. Termina;
  BEGIN
    Bala. Termina;
  END;

  CONSTRUCTOR OGrupoInvasor. Prepara;

    VAR
      IndiceInvasor : Byte;
      NuevaX : Word;

  BEGIN
    CambioX := 1;
    NuevaX := 0;

    FOR IndiceInvasor := 1 TO 10 DO
      WITH Invasores [IndiceInvasor] DO
      BEGIN
        Prepara (NuevaX, 0, Random (4) + 2);
        Inc (NuevaX, 20);
      END;

    Muertos := False;
  END;

  PROCEDURE OGrupoInvasor. Transforma;

    VAR
      IndiceInvasor : Byte;
      NuevaX : Word;
      NuevaY : Word;
      CuentaMuertos : Byte;

  BEGIN
    NuevaX := Invasores [1]. X;
    NuevaY := Invasores [1]. Y;

    IF Random (30) = 29 THEN Inc (NuevaY);

    IF CambioX > 0 THEN
    BEGIN

      IF NuevaX < 101 THEN
        Inc (NuevaX)
      ELSE
        CambioX := -1;

    END ELSE
      IF CambioX < 0 THEN
      BEGIN

        IF NuevaX > 0 THEN
          Dec (NuevaX)
        ELSE
          CambioX := 1;

      END;

    CuentaMuertos := 0;

    FOR IndiceInvasor := 1 TO 10 DO
      WITH Invasores [IndiceInvasor] DO
      BEGIN
        IF Muerto THEN Inc (CuentaMuertos);
        X := NuevaX;
        Y := NuevaY;
        Transforma;
        Inc (NuevaX, 20);
      END;

    IF CuentaMuertos = 10 THEN Muertos := True;
  END;

  PROCEDURE OGrupoInvasor. DetectaChoque;

    VAR
      IndiceInvasor : Byte;

  BEGIN
    FOR IndiceInvasor := 1 TO 10 DO
      Invasores [IndiceInvasor]. DetectaChoque;
  END;

  PROCEDURE OGrupoInvasor. Muestra;

    VAR
      IndiceInvasor : Byte;

  BEGIN
    FOR IndiceInvasor := 1 TO 10 DO
      Invasores [IndiceInvasor]. Muestra;
  END;

  DESTRUCTOR OGrupoInvasor. Termina;

    VAR
      IndiceInvasor : Byte;

  BEGIN
    FOR IndiceInvasor := 1 TO 10 DO
      Invasores [IndiceInvasor]. Termina;
  END;

  PROCEDURE PreparaObjetos;
  BEGIN
    Artilleria. Prepara;
    GrupoInvasor. Prepara;
  END;

  PROCEDURE TerminaObjetos;
  BEGIN
    GrupoInvasor. Termina;
    Artilleria. Termina;
  END;

{ OBJETOS                                                                    }
{----------------------------------------------------------------------------}

  VAR
    FinCiclo : Boolean;
    Ganador : Boolean;
    Aterrizaron : Boolean;

  PROCEDURE Prepara;
  BEGIN
    Randomize;
    WriteLn ('INVASORES');
    WriteLn ('(c) 1996 Marco A. Alvarado');
    WriteLn;
    WriteLn ('Correo Electrnico:');
    WriteLn ('  marcalva@sol.racsa.co.cr');
    WriteLn;
    WriteLn ('Apartado Postal:');
    WriteLn ('  Marco A. Alvarado');
    WriteLn ('  318-1002 Estudiantes');
    WriteLn ('  San Jos, Costa Rica');
    WriteLn;
    WriteLn ('Durante el juego use las flechas, la barra, y [Esc] para salir.');
    WriteLn;
    WriteLn ('Listo! Presione [Esc] para continuar.');
    WriteLn;
    FinCiclo := False;
    Ganador := True;
    Aterrizaron := False;
    PreparaTeclado;
    repeat until teclaesc;
    repeat until not teclaesc;
    PreparaVideo;
    PreparaObjetos;
  END;

  PROCEDURE Cicla;
  BEGIN
    Barrera. Muestra;

    REPEAT

      IF (NOT GrupoInvasor. Muertos) AND (GrupoInvasor. Invasores [1]. Y <
        140) THEN
      BEGIN

        IF NOT Artilleria. Muerta THEN
        BEGIN
          Artilleria. Transforma;
          Artilleria. DetectaChoque;
          Artilleria. Muestra;
        END ELSE
          Ganador := False;

        GrupoInvasor. Transforma;
        GrupoInvasor. DetectaChoque;
        GrupoInvasor. Muestra;
      END ELSE
        IF GrupoInvasor. Invasores [1]. Y = 140 THEN
        BEGIN
          Ganador := False;
          Aterrizaron := True;
        END;

      IF TeclaEsc THEN FinCiclo := True;
      Delay(10);
    UNTIL FinCiclo;

  END;

  PROCEDURE Termina;
  BEGIN
    TerminaObjetos;
    TerminaVideo;
    TerminaTeclado;
    IF Ganador THEN
      WriteLn ('Ganaste esta vez terrcola, pero... volveremos!')
    ELSE
      IF Aterrizaron THEN
        WriteLn ('Ahora nos apoderaremos de tu planeta.')
      ELSE
        WriteLn ('Ja, ja, ja... perdiste.');
    WriteLn;
  END;

BEGIN
  Prepara;
  Cicla;
  Termina;
END.

{ INVASORES                                                                  }
{----------------------------------------------------------------------------}